#define INDBGMEM_C
#include <stdio.h>
#include <stdlib.h>

#define MEM_LOGOUTOFMEM
#define MEM_TRACK

#include <error.h>
#include <dbgmem.h>

/* make these dividable by 4 for 32bit alignment */
#define MEM_STARTBLOCKSIZE 4
#define MEM_ENDBLOCKSIZE 4

#define MEM_DEADMEM ((unsigned char) 0xAA)
#define MEM_WASALLOCATED ((unsigned int) 0xDEAD)


struct memblock {
  int allocated;
  size_t size;
  char *file;
  int line;
#ifdef MEM_TRACK
  struct memblock *next;
#endif
  unsigned char pad[MEM_STARTBLOCKSIZE];
};

#ifdef MEM_TRACK
 static struct memblock *DbgAllocatedList=0;
 int DbgNumAllocated=0;
#endif


static int DbgVerifyBefore(struct memblock *blk)
{
  int overwrite;
  int i;

  overwrite=0;
  for (i=0;i<MEM_STARTBLOCKSIZE;i++){
    if (blk->pad[i]!=MEM_DEADMEM){
      overwrite=1;
    }
  }
  return overwrite;
}

static int DbgVerifyAfter(struct memblock *blk)
{
  int i;
  unsigned char *cptr;
  int overwrite;
  
  cptr=(((unsigned char *)blk)+sizeof(struct memblock)+blk->size);
  overwrite=0;
  for (i=0;i<MEM_ENDBLOCKSIZE;i++){
    if (cptr[i]!=MEM_DEADMEM){
      overwrite=1;
    }
  }
  return overwrite;
}

void *DbgMalloc(size_t size,char *file,int line)
{
  void *ptr;
  struct memblock *blk;
  unsigned char *cptr;
  int i;

  ptr=malloc(size+sizeof(struct memblock)+MEM_ENDBLOCKSIZE);
  if (ptr==0) {
#ifdef MEM_LOGOUTOFMEM
    _ErrWarning("Out of memory in malloc",file,line);
#endif
    return ptr;
  }
  blk=ptr;
  blk->file=file;
  blk->line=line;
  blk->allocated=MEM_WASALLOCATED;
  blk->size=size;
#ifdef MEM_TRACK
  blk->next=DbgAllocatedList;
  DbgAllocatedList=blk;
#endif
  DbgNumAllocated++;
  for (i=0;i<MEM_STARTBLOCKSIZE;i++) {
    blk->pad[i]=MEM_DEADMEM;
  }
  cptr=((unsigned char *) ptr)+sizeof(struct memblock)+size;
  for (i=0;i<MEM_ENDBLOCKSIZE;i++) {
    cptr[i]=MEM_DEADMEM;
  }
  return ((unsigned char *) ptr)+sizeof(struct memblock);

}

void *DbgCalloc(size_t size,char *file,int line)
{
  void *ptr;
  char *cptr;
  int i;

  ptr=DbgMalloc(size,file,line);
  if (ptr==0)
    return ptr;
  cptr=ptr;
  for(i=0;i<size;i++) {
   cptr[i]=0;
  }
  return ptr;
}

void DbgFree(void *ptr,char *file,int line)
{
  void *real;
  struct memblock *blk,*bptr;
  struct memblock **lastnext;
  int i;

  if (ptr==0) {
    _ErrError("Freeing a zero pointer",file,line);
  }
  real=(void *) (((unsigned char *) ptr)-sizeof(struct memblock));
  blk=real;
#ifdef MEM_TRACK
  lastnext=&DbgAllocatedList;
  bptr=DbgAllocatedList;
  i=0;
  while (bptr!=0) {
    if (bptr==blk) {
      *lastnext=bptr->next;
      break;
    }
    lastnext=&(bptr->next);
    bptr=bptr->next;
    if (++i>DbgNumAllocated)
      _ErrError("Internal error: Memory list corrupt",__FILE__,__LINE__);
  }
  if (bptr==0)
    _ErrError("Freeing an already freed or not allocated block",file,line);
  /* find block in list */
#else
  if (blk->allocated!=MEM_WASALLOCATED) {
    _ErrError("Freeing an already freed or not allocated block",file,line);
  }
#endif
  
  DbgNumAllocated--;

  blk->allocated=0;
  
  if (DbgVerifyBefore(blk)) {
    _ErrError("The memory block allocated here was written before.",
	      blk->file,blk->line);
  }
  if (DbgVerifyAfter(blk)) {
    _ErrError("The memory block allocated here was written after.",
	      blk->file,blk->line);
  }

  return;
}

void *DbgReAlloc(void *ptr,size_t size,char *file,int line)
{
  void *newptr;
  int newsize;
  struct memblock *blk;
  struct memblock *p;
  int found,i;

  blk=(struct memblock *) ptr - 1;

  newptr=DbgMalloc(size,file,line);
  if (newptr==0)
    return 0;
  if (ptr==0) {
    return newptr;
  }

#ifdef MEM_TRACK
  found=0;
  p=DbgAllocatedList;
  i=0;
  while (p!=0) {
    if (p==blk) {
      found=1;
      break;
    }
    p=p->next;
    if (++i>DbgNumAllocated) 
      _ErrError("Internal error: Memory list corrupt",__FILE__,__LINE__);
  }
  if (!found)
    _ErrError("Passed non allocated pointer to realloc()",file,line);
#else
  if (blk->allocated!=MEM_WASALLOCATED){
    _ErrError("Passed non allocated pointer to realloc()",file,line);
  }
#endif

  newsize=blk->size;
  newsize= (size<newsize) ? size : newsize;
  memcpy(newptr,ptr,newsize);
  DbgFree(ptr,file,line);
  return newptr;
}

void DbgMemStat(void)
{
  struct memblock *bptr;
  int i,ok;

  printf("Number of allocated blocks: %d\n",DbgNumAllocated);

#ifdef MEM_TRACK
  bptr=DbgAllocatedList;
  i=0;
  while (bptr!=0) {
    ok=1;
    printf("%s:%d (",bptr->file,bptr->line);
    if (bptr->size<32000) {
      printf("%d bytes)",bptr->size);
    } else if (bptr->size<1024*1024) {
      printf("%d Kb)",bptr->size/1024);
    } else {
      printf("%d Mb)",bptr->size/(1024*1024));
    }  

    if (bptr->allocated!=MEM_WASALLOCATED) {
      printf(" Not allocated");
      ok=0;
    } 
    if (DbgVerifyBefore(bptr)) {
      printf(" Written before");
      ok=0;
    }
    if (DbgVerifyAfter(bptr)) {
      printf(" Written after");
      ok=0;
    }
    if (ok) {
      printf(" ok");
    }
    printf("\n");
    bptr=bptr->next;
    if (++i>DbgNumAllocated) {
      _ErrError("Internal error: Memory list corrupt",__FILE__,__LINE__);
      break;
    }
  }
#endif
}
